/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
#include <limits.h>
#include <memory.h>
#include <string.h>
#include <stropts.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
               
#include <dnpap.h>
#include <block.h>
#include <config.h>
#include <message.h>

#include "nw.h"
#include "nwdlpi.h"

 
/* nwdlpi.c - NwDlpi module functions */

            
NW_DG_IFACE DlpiDgIface = { "dlpi", 
                            NwDlpiInit,   NwDlpiExit, 
                            NwDlpiAccept, NwDlpiConnect, 
                            NwDlpiSend,   NwDlpiSendTo,
                            NwDlpiClose };


static NwDlpiDg *NwDlpiDgList = NULL;


static INT dlpisocket;

static BOOLEAN NwDlpiCloseDg(NwDlpiDg *dg);

static BOOLEAN NwDlpiDgAddList(NwDlpiDg *dg);
static BOOLEAN NwDlpiDgRemoveList(NwDlpiDg *dg);

static NwDlpiDg* NewNwDlpiDg(CHAR *interface, ULONG bufsize);
static VOID DelNwDlpiDg(NwDlpiDg *dg);

static BOOLEAN NwDlpiRead(NwDlpiDg *dg);

static BOOLEAN NwDlpiSetup(NwDlpiDg *dg, CHAR *interface, ULONG flags, ULONG bufsize, ULONG snaplen, TIMEVAL *timeout);

static int NwDlpiBlockCallback(BLOCK *blk, int reason, VOID *param);



/* Initializes the NwDlpi module
 * Initializes the NwDlpi messages and opens a socket for general use
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwDlpiInit(VOID)
{       
    MessageConfig(NWDLPI_ERROR, "NwDlpi");

	if (geteuid() != 0)
	{
        DnpapMessage(DMC_WARNING, NWDLPI_ROOT, "nwdlpi: must be root to use dlpi features");
		return TRUE;
	}
	
    if ((dlpisocket = open(NWDLPIDEVICE, O_WRONLY)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_OPEN, "nwdlpi: open failed");
        return FALSE;
    }

    return TRUE;
}


/* Exits the NwDlpi module
 * Closes all opened sockets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwDlpiExit(VOID)
{             
BOOLEAN rc = TRUE;
NwDlpiDg *dg;

    for (dg = NwDlpiDgList; dg != NULL; dg = NwDlpiDgList)
        rc &= NwDlpiClose((NW_DG *)dg);

    if (dlpisocket != 0)
        close(dlpisocket);        
        
    return rc;
}

        
/* Opens a socket for receiving promiscuous packets
 * Opens the specified interface with capacity 'bufsize' and snapshot
 * length 'snaplen'. 'bufsize' space is automatically allocated.
 * Set 'snaplen' to 0 for no snapshot at all
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG *NwDlpiAccept(char *addr           /* accept specification: 'dlpi:(char*)interface:(ulong)bufsize:(ulong)snaplen' */, 
                   NW_DG_RECV_CALL recv  /* receive callback; parameter addr is formatted as follows: 'dlpi:(char*)interface:(ulong)bufsize:(ulong)snaplen' */, 
                   VOID *param           /* extra parameter you want to be forwarded to recv() */)
{   
NwDlpiDg *dg;
CHAR interface[NWDLPI_SIZE_INTERFACE];
ULONG bufsize, snaplen;
ULONG flags;
TIMEVAL timeout;
CHAR *p;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameter list");
        return NULL;
    }   
    
    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
	if (sscanf(p, ":%lu:%lu", &bufsize, &snaplen) != 2)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameters");
	    return NULL;
	}
	                           
    /* set buffer timeout at 1/10 of a second */
	timeout.tv_sec = 0;
	timeout.tv_usec = 100000UL;

	if ((dg = NewNwDlpiDg(interface, bufsize)) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWDLPI_NEWDLPI, "nwdlpi: memory allocation failed");
		return NULL;
    }

    if ((dg->DlpiInfo.Socket = open(NWDLPIDEVICE, O_RDONLY)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_OPEN, "nwdlpi: open failed");
    	NwDlpiCloseDg(dg);
        return NULL;
    }

    if (NwDlpiSetup(dg, interface, flags, bufsize, snaplen, &timeout) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SETUP, "nwdlpi: setup failed");
    	NwDlpiCloseDg(dg);
        return NULL;
    }
    
    dg->Block->sckt = dg->DlpiInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwDlpiBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_BLOCK, "nwdlpi: block registration failed");
    	NwDlpiCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &DlpiDgIface;                                
    sprintf(dg->NwInfo.addr, "dlpi:%s:%lu:%lu", interface, bufsize, snaplen);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;    
    
	strcpy(dg->DlpiInfo.Interface, interface);
	dg->DlpiInfo.BufSize = bufsize;
	dg->DlpiInfo.SnapLen = snaplen;
	dg->DlpiInfo.Timeout.tv_sec = timeout.tv_sec;
	dg->DlpiInfo.Timeout.tv_usec = timeout.tv_usec;
	dg->DlpiInfo.Flags = flags;
	dg->DlpiInfo.Pos = 0;
	dg->DlpiInfo.End = 0;
             
    if (NwDlpiDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_ADDLIST, "nwdlpi: could not add to list");
    	NwDlpiCloseDg(dg);
		return NULL;
	}
    
    if (ioctl(dg->DlpiInfo.Socket, I_FLUSH, (VOID *)FLUSHR) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: read flush failed");
    	NwDlpiCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


/* Opens a socket for receiving and sending packets from and to a specific host
 * Opens the specified interface with capacity 'bufsize' and snapshot
 * length 'snaplen'. 'bufsize' space is automatically allocated.
 * Set 'snaplen' to 0 for no snapshot at all. The Ethernet address is recorded 
 * and the receive callback is only called for packets to or from this address
 * Returns a NW datagram pointer or NULL for failure
 */
NW_DG  *NwDlpiConnect(char *addr            /* connect specification: 'dlpi:(char*)interface:(ulong)bufsize:(ulong)snaplen:(int[6])ethernet' */, 
                     NW_DG_RECV_CALL recv  /* receive callback; parameter addr is formatted as follows: 'dlpi:(char*)interface:(ulong)bufsize:(ulong)snaplen' */, 
                     VOID *param           /* extra parameter you want to forward to recv() */)
{   
NwDlpiDg *dg;
CHAR interface[NWDLPI_SIZE_INTERFACE];
ULONG bufsize, snaplen;
ULONG flags;
TIMEVAL timeout;
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;

    if ((addr = strchr(addr, ':')) == NULL || (p = strchr(++addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameter list");
        return NULL;
    }   
    
    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
	if (sscanf(p, ":%lu:%lu:%u.%u.%u.%u.%u.%u", &bufsize, &snaplen, &a1, &a2, &a3, &a4, &a5, &a6) != 8)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameters");
	    return NULL;
	}
	
    /* set buffer timeout at 1/10 of a second */
	timeout.tv_sec = 0;
	timeout.tv_usec = 100000UL;

	if ((dg = NewNwDlpiDg(interface, bufsize)) == NULL)    
    {
        DnpapMessage(DMC_ERROR, NWDLPI_NEWDLPI, "nwdlpi: memory allocation failed");
		return NULL;
    }

    if ((dg->DlpiInfo.Socket = open(NWDLPIDEVICE, O_RDWR)) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_OPEN, "nwdlpi: open failed");
    	NwDlpiCloseDg(dg);
        return NULL;
    }

    if (NwDlpiSetup(dg, interface, flags, bufsize, snaplen, &timeout) == FALSE)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SETUP, "nwdlpi: setup failed");
    	NwDlpiCloseDg(dg);
        return NULL;
    }
    
    dg->Block->sckt = dg->DlpiInfo.Socket;
    
    if (BlockRegister(dg->Block, dg, NwDlpiBlockCallback) != 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_BLOCK, "nwdlpi: block registration failed");
    	NwDlpiCloseDg(dg);
		return NULL;
    }

    dg->NwInfo.iface = &DlpiDgIface;                                
    sprintf(dg->NwInfo.addr, "dlpi:%s:%lu:%lu:%u.%u.%u.%u.%u.%u", 
                             interface, bufsize, snaplen, a1, a2, a3, a4, a5, a6);
	dg->NwInfo.recv = recv;
    dg->NwInfo.parm = param;    
	
	strcpy(dg->DlpiInfo.Interface, interface);
    dg->DlpiInfo.Address[0] = a1;
    dg->DlpiInfo.Address[1] = a2;
    dg->DlpiInfo.Address[2] = a3;
    dg->DlpiInfo.Address[3] = a4;
    dg->DlpiInfo.Address[4] = a5;
    dg->DlpiInfo.Address[5] = a6;
	dg->DlpiInfo.BufSize = bufsize;
	dg->DlpiInfo.SnapLen = snaplen;
	dg->DlpiInfo.Timeout.tv_sec = timeout.tv_sec;
	dg->DlpiInfo.Timeout.tv_usec = timeout.tv_usec;
	dg->DlpiInfo.Flags = flags;
	dg->DlpiInfo.Pos = 0;
	dg->DlpiInfo.End = 0;
             
    if (NwDlpiDgAddList(dg) == FALSE)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_ADDLIST, "nwdlpi: could not add to list");
    	NwDlpiCloseDg(dg);
		return NULL;
	}
    
    if (ioctl(dg->DlpiInfo.Socket, I_FLUSH, (VOID *)FLUSHR) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: read flush failed");
    	NwDlpiCloseDg(dg);
		return NULL;
	}
    
    return (NW_DG *)dg;
}


/* Closes a NwDlpi datagram socket
 * Closes the specified NwDlpi datagram socket
 * Returns TRUE for success and FALSE for failure
 */
BOOLEAN NwDlpiClose(NW_DG *dg   /* an existing NwDlpi datagram socket */)
{                   
NwDlpiDg *dlpidg = (NwDlpiDg *)dg;

    NwDlpiDgRemoveList(dlpidg);
    NwDlpiCloseDg(dlpidg);

    return TRUE;
}


BOOLEAN NwDlpiCloseDg(NwDlpiDg *dg)
{
    if (dg->Block != NULL)    
        BlockDeRegister(dg->Block);
    
    if (dg->DlpiInfo.Socket != 0)
        close(dg->DlpiInfo.Socket);
        
    DelNwDlpiDg(dg);  
    
    return TRUE;
}    
    

/* Sends a packet over an existing NW datagram
 * The destination address is checked against the recorded Ethernet address
 * to allow only packets that are sent to that address
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwDlpiSend(NW_DG *dg    /* valid NW datagram */, 
                  BYTE *frame  /* packet */, 
                  INT len      /* packet length */)
{
NwDlpiDg *dlpidg = (NwDlpiDg *) dg;	
SOCKADDR sa; 
STRBUF ctl, data;
INT llhsize = sizeof(sa.sa_data);
BYTE noaddr[NWDLPI_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    if (len > NWDLPI_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SEND, "nwdlpi: packet too long");
        return FALSE;
    }

    if (memcmp(dlpidg->DlpiInfo.Address, noaddr, NWDLPI_SIZE_ADDRESS) == 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SEND, "nwdlpi: can not send over accept socket");
        return FALSE;
    }
            
    sa.sa_family = AF_UNSPEC;
    memcpy(frame, dlpidg->DlpiInfo.Address, NWDLPI_SIZE_ADDRESS);
    memcpy(sa.sa_data, frame, llhsize);

    ctl.maxlen = ctl.len = sizeof(sa);
    ctl.buf = (BYTE *)&sa;
    data.maxlen = data.len = len - llhsize;
    data.buf = frame + llhsize;
    if (putmsg(dlpidg->DlpiInfo.Socket, &ctl, &data, 0) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SEND, "nwdlpi: putmsg failed");
        return FALSE;
    }

    if (ioctl(dlpidg->DlpiInfo.Socket, I_FLUSH, (VOID *)FLUSHW) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: write flush failed");
        return FALSE;
    }
        
    return TRUE;
}


/* Sends a packet to a specified address
 * This function can be used to sent one-shot only packets
 * Returns TRUE or FALSE for success or failure
 */
BOOLEAN NwDlpiSendTo(NW_DG *dg    /* not supported, set to NULL */, 
                    char *addr   /* address specification: 'dlpi:(char*)interface[:(int[6])ethernet]' */, 
                    BYTE *frame  /* packet */, 
                    INT len      /* packet length */)
{
STRIOCTL si;
IFREQ ifr;
SOCKADDR sa; 
STRBUF ctl, data;
INT llhsize = sizeof(sa.sa_data);
CHAR interface[NWDLPI_SIZE_INTERFACE];
CHAR *p;
UINT a1, a2, a3, a4, a5, a6;
BYTE address[NWDLPI_SIZE_ADDRESS];
         
    if (dg != NULL)
        DnpapMessage(DMC_WARNING, NWDLPI_SEND, "nwdlpi: sendto over existing connection has no effect");
    
    if (dlpisocket == 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_IDLPI, "nwdlpi: not initialized");
        return FALSE;
    }
    
    if (len > NWDLPI_SIZE_PACKET)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SEND, "nwdlpi: packet too long");
        return FALSE;
    }

    if ((addr = strchr(addr, ':')) == NULL)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameter list");
        return FALSE;
    }   
    									  
	if ((p = strchr(++addr, ':')) == NULL)
	{
		p = addr+strlen(addr);
	}
	else
	{
		if (sscanf(p, ":%u.%u.%u.%u.%u.%u", &a1, &a2, &a3, &a4, &a5, &a6) != 6)
		{
	        DnpapMessage(DMC_ERROR, NWDLPI_PARAM, "nwdlpi: invalid parameters");
		    return FALSE;
		}
		
	    address[0] = a1;	
	    address[1] = a2;	
	    address[2] = a3;	
	    address[3] = a4;	
	    address[4] = a5;	
	    address[5] = a6;	
		
	    memcpy(frame, address, NWDLPI_SIZE_ADDRESS);
    }

    strncpy(interface, addr, p-addr);
    interface[p-addr] = '\0';
    
    memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
    strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));
    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';
    si.ic_timout = INFTIM;
    si.ic_cmd = NIOCBIND;
    si.ic_len = sizeof(ifr);
    si.ic_dp = (VOID *)&ifr;
    if (ioctl(dlpisocket, I_STR, (VOID *)&si) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: interface bind failed");
		return FALSE;
	}
    
    sa.sa_family = AF_UNSPEC;
    memcpy(sa.sa_data, frame, llhsize);

    ctl.maxlen = ctl.len = sizeof(sa);
    ctl.buf = (BYTE *)&sa;
    data.maxlen = data.len = len - llhsize;
    data.buf = frame + llhsize;
    if (putmsg(dlpisocket, &ctl, &data, 0) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_SEND, "nwdlpi: putmsg failed");
        return FALSE;
    }

    if (ioctl(dlpisocket, I_FLUSH, (VOID *)FLUSHW) < 0)
    {
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: write flush failed");
        return FALSE;
    }
        
    return TRUE;
}


int NwDlpiBlockCallback(BLOCK *blk, int reason, VOID *param)
{
NwDlpiDg *dg = (NwDlpiDg *) param;	
	
	switch (reason)
	{
	case BLOCKEVENT_READ:
		if (NwDlpiRead(dg) != TRUE)
			DnpapMessage(DMC_ERROR, NWDLPI_READ, "nwdlpi: read failed");
		break;
	case BLOCKEVENT_WRITE:
		DnpapMessage(DMC_ERROR, NWDLPI_WEVENT, "nwdlpi: write event");
		break;
	case BLOCKEVENT_EXCEPT:
		DnpapMessage(DMC_ERROR, NWDLPI_EEVENT, "nwdlpi: except event");
		break;
	default:
		DnpapMessage(DMC_ERROR, NWDLPI_NEVENT, "nwdlpi: unknown block event");
		return 1;
	}
	
	return 0;
}


BOOLEAN NwDlpiRead(NwDlpiDg *dg)
{
BYTE *cp;
DLPIBUFHDR *hp;
ULONG msglen;
BYTE noaddr[NWDLPI_SIZE_ADDRESS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    if (!(dg->DlpiInfo.Pos < dg->DlpiInfo.End))
    {
    	if ((dg->DlpiInfo.End = read(dg->DlpiInfo.Socket, dg->DlpiInfo.Buf, dg->DlpiInfo.BufSize)) < 0)
    		return FALSE;
            
    	dg->DlpiInfo.Pos = 0;
    }
		
	while (dg->DlpiInfo.Pos < dg->DlpiInfo.End)
	{
		cp = dg->DlpiInfo.Buf+dg->DlpiInfo.Pos;
		
		hp = (DLPIBUFHDR *)cp;      
		cp += sizeof(*hp);
	    msglen = ((IFLEN *)(cp+sizeof(IFTIME)+sizeof(IFDROPS)))->nh_pktlen;
		if (dg->DlpiInfo.SnapLen > 0 && msglen > dg->DlpiInfo.SnapLen)
			msglen = dg->DlpiInfo.SnapLen;
		
		dg->DlpiInfo.Pos += hp->nhb_totlen;

        if (memcmp(dg->DlpiInfo.Address, noaddr, NWDLPI_SIZE_ADDRESS) != 0 &&
            memcmp(dg->DlpiInfo.Address, cp, NWDLPI_SIZE_ADDRESS) != 0 &&
            memcmp(dg->DlpiInfo.Address, cp+NWDLPI_SIZE_ADDRESS, NWDLPI_SIZE_ADDRESS) != 0)
            continue;
        
		if (dg->NwInfo.recv != NULL)
			dg->NwInfo.recv((NW_DG *)dg, dg->NwInfo.addr, cp, msglen, dg->NwInfo.parm);
	}
	
	return TRUE;
}


BOOLEAN NwDlpiDgAddList(NwDlpiDg *dg)
{
    if (NwDlpiDgList != NULL)
		NwDlpiDgList->Prev = dg;
	
    dg->Next = NwDlpiDgList;
    dg->Prev = NULL;
    
    NwDlpiDgList = dg;
        
    return TRUE;
}


BOOLEAN NwDlpiDgRemoveList(NwDlpiDg *dg)
{   
    if (dg->Prev != NULL)
        dg->Prev->Next = dg->Next;
    else
        NwDlpiDgList = dg->Next;
        
    if (dg->Next != NULL)
        dg->Next->Prev = dg->Prev;
        
    return TRUE;
}


NwDlpiDg* NewNwDlpiDg(CHAR *interface, ULONG bufsize)
{
NwDlpiDg *dg;

	if ((dg = DnpapMalloc(sizeof(NwDlpiDg))) == NULL)
		return NULL;
	
	dg->Block = NULL;
    
    dg->NwInfo.iface = NULL;
	dg->NwInfo.recv = NULL;
    dg->NwInfo.parm = NULL;    
    
    dg->DlpiInfo.Interface = NULL;
    memset(dg->DlpiInfo.Address, 0, NWDLPI_SIZE_ADDRESS);
    dg->DlpiInfo.Buf = NULL;
	dg->DlpiInfo.Socket = 0;		
	dg->DlpiInfo.BufSize = 0;
	dg->DlpiInfo.SnapLen = 0;
	dg->DlpiInfo.Timeout.tv_sec = 0;
	dg->DlpiInfo.Timeout.tv_usec = 0;
	dg->DlpiInfo.Flags = 0;
	dg->DlpiInfo.Pos = 0;
	dg->DlpiInfo.End = 0;
    
    dg->Next = NULL;
    dg->Prev = NULL;
	
    if ((dg->Block = DnpapMalloc(sizeof(BLOCK))) == NULL)
		return NULL;
    
	if ((dg->DlpiInfo.Interface = DnpapMalloc(strlen(interface)+1)) == NULL)
    {
    	DnpapFree(dg->Block);
		return NULL;
	}
	
	if ((dg->DlpiInfo.Buf = DnpapMalloc(bufsize)) == NULL)
    {
    	DnpapFree(dg->Block);
    	DnpapFree(dg->DlpiInfo.Interface);
		return NULL;
	}
	
	return dg;
}


VOID DelNwDlpiDg(NwDlpiDg *dg)
{
	DnpapFree(dg->Block);
	
    DnpapFree(dg->DlpiInfo.Interface);
    DnpapFree(dg->DlpiInfo.Buf);
    
    DnpapFree(dg);
}


BOOLEAN NwDlpiSetup(NwDlpiDg *dg, CHAR *interface, ULONG flags, ULONG bufsize, ULONG snaplen, TIMEVAL *timeout)
{
STRIOCTL si;
IFREQ ifr;
 
    flags |= NWDLPI_DEFAULT_FLAGS;
    
    /* arrange to get discrete messages from the stream */
    if (ioctl(dg->DlpiInfo.Socket, I_SRDOPT, (VOID *)RMSGD) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: discrete message option set failed");
		return FALSE;
	}
    
    /* setting of options are allowed to take a long time */      
    si.ic_timout = INFTIM;
                   
    /* push and configure the filtering module */
    if (ioctl(dg->DlpiInfo.Socket, I_PUSH, NWDLPIFILTER) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_PUSH, "nwdlpi: 'pfmod' push failed");
		return FALSE;
	}
    
    /* push and configure the buffering module */
    if (ioctl(dg->DlpiInfo.Socket, I_PUSH, NWDLPIBUFFER) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_PUSH, "nwdlpi: 'bufmod' push failed");
		return FALSE;
	}
    
    si.ic_cmd = NIOCSCHUNK;
    si.ic_len = sizeof(bufsize);
    si.ic_dp = (VOID *)&bufsize;
    if (ioctl(dg->DlpiInfo.Socket, I_STR, (VOID *)&si) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: chunksize option set failed");
		return FALSE;
	}

	/* timeout settings must come after bufsize settings:
	 * if timeout == 0 then bufsize is reset too
	 */
    si.ic_cmd = NIOCSTIME;
    si.ic_len = sizeof(*timeout);
    si.ic_dp = (VOID *)timeout;
    if (ioctl(dg->DlpiInfo.Socket, I_STR, (VOID *)&si) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: timeout option set failed");
		return FALSE;
	}
    
    memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));
    strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));
    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';
    si.ic_cmd = NIOCBIND;
    si.ic_len = sizeof(ifr);
    si.ic_dp = (VOID *)&ifr;
    if (ioctl(dg->DlpiInfo.Socket, I_STR, (VOID *)&si) < 0)
	{
        DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: interface bind failed");
		return FALSE;
	}
    
    if (snaplen > 0)
    {
        si.ic_cmd = NIOCSSNAP;
        si.ic_len = sizeof(snaplen);
        si.ic_dp = (VOID *)&snaplen;
        if (ioctl(dg->DlpiInfo.Socket, I_STR, (VOID *)&si) < 0)
    	{
            DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: snapshot length option set failed");
    		return FALSE;
    	}
    }
                                   
    if (flags != 0)
    {
        si.ic_cmd = NIOCSFLAGS;
        si.ic_len = sizeof(flags);
        si.ic_dp = (VOID *)&flags;
        if (ioctl(dg->DlpiInfo.Socket, I_STR, (VOID *)&si) < 0)
    	{
            DnpapMessage(DMC_ERROR, NWDLPI_IOCTL, "nwdlpi: flags set failed");
    		return FALSE;
    	}
    }
    
    return TRUE;
}
